import { TextService, CanvasConfig, Camera, ContextService, SceneGraphService, RenderingContext, DisplayObjectPool, Renderable, SCENE_GRAPH_EVENT, AABB, getEuler, fromRotationTranslationScale, SHAPE, world, container, RenderingPluginContribution } from '@antv/g';
import RBush from 'rbush';
export { default as RBush } from 'rbush';
import { Component } from '@antv/g-ecs';

var NAMED_TAG = "named";
var INJECT_TAG = "inject";
var TAGGED = "inversify:tagged";
var TAGGED_PROP = "inversify:tagged_props";
var PARAM_TYPES = "inversify:paramtypes";
var DESIGN_PARAM_TYPES = "design:paramtypes";

var idCounter = 0;
function id() {
    return idCounter++;
}

var DUPLICATED_INJECTABLE_DECORATOR = "Cannot apply @injectable decorator multiple times.";
var DUPLICATED_METADATA = "Metadata key was used more than once in a parameter:";
var UNDEFINED_INJECT_ANNOTATION = function (name) {
    return "@inject called with undefined this could mean that the class " + name + " has " +
        "a circular dependency problem. You can use a LazyServiceIdentifer to  " +
        "overcome this limitation.";
};
var INVALID_DECORATOR_OPERATION = "The @inject @multiInject @tagged and @named decorators " +
    "must be applied to the parameters of a class constructor or a class property.";

var Metadata = (function () {
    function Metadata(key, value) {
        this.key = key;
        this.value = value;
    }
    Metadata.prototype.toString = function () {
        if (this.key === NAMED_TAG) {
            return "named: " + this.value.toString() + " ";
        }
        else {
            return "tagged: { key:" + this.key.toString() + ", value: " + this.value + " }";
        }
    };
    return Metadata;
}());

function tagParameter(annotationTarget, propertyName, parameterIndex, metadata) {
    var metadataKey = TAGGED;
    _tagParameterOrProperty(metadataKey, annotationTarget, propertyName, metadata, parameterIndex);
}
function tagProperty(annotationTarget, propertyName, metadata) {
    var metadataKey = TAGGED_PROP;
    _tagParameterOrProperty(metadataKey, annotationTarget.constructor, propertyName, metadata);
}
function _tagParameterOrProperty(metadataKey, annotationTarget, propertyName, metadata, parameterIndex) {
    var paramsOrPropertiesMetadata = {};
    var isParameterDecorator = (typeof parameterIndex === "number");
    var key = (parameterIndex !== undefined && isParameterDecorator) ? parameterIndex.toString() : propertyName;
    if (isParameterDecorator && propertyName !== undefined) {
        throw new Error(INVALID_DECORATOR_OPERATION);
    }
    if (Reflect.hasOwnMetadata(metadataKey, annotationTarget)) {
        paramsOrPropertiesMetadata = Reflect.getMetadata(metadataKey, annotationTarget);
    }
    var paramOrPropertyMetadata = paramsOrPropertiesMetadata[key];
    if (!Array.isArray(paramOrPropertyMetadata)) {
        paramOrPropertyMetadata = [];
    }
    else {
        for (var _i = 0, paramOrPropertyMetadata_1 = paramOrPropertyMetadata; _i < paramOrPropertyMetadata_1.length; _i++) {
            var m = paramOrPropertyMetadata_1[_i];
            if (m.key === metadata.key) {
                throw new Error(DUPLICATED_METADATA + " " + m.key.toString());
            }
        }
    }
    paramOrPropertyMetadata.push(metadata);
    paramsOrPropertiesMetadata[key] = paramOrPropertyMetadata;
    Reflect.defineMetadata(metadataKey, paramsOrPropertiesMetadata, annotationTarget);
}

function inject(serviceIdentifier) {
    return function (target, targetKey, index) {
        if (serviceIdentifier === undefined) {
            throw new Error(UNDEFINED_INJECT_ANNOTATION(target.name));
        }
        var metadata = new Metadata(INJECT_TAG, serviceIdentifier);
        if (typeof index === "number") {
            tagParameter(target, targetKey, index, metadata);
        }
        else {
            tagProperty(target, targetKey, metadata);
        }
    };
}

var __spreadArrays = (undefined && undefined.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};

var __spreadArrays$1 = (undefined && undefined.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};

var __awaiter = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (undefined && undefined.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};

var ContainerModule = (function () {
    function ContainerModule(registry) {
        this.id = id();
        this.registry = registry;
    }
    return ContainerModule;
}());

function injectable() {
    return function (target) {
        if (Reflect.hasOwnMetadata(PARAM_TYPES, target)) {
            throw new Error(DUPLICATED_INJECTABLE_DECORATOR);
        }
        var types = Reflect.getMetadata(DESIGN_PARAM_TYPES, target) || [];
        Reflect.defineMetadata(PARAM_TYPES, types, target);
        return target;
    };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _createSuper(Derived) {
  var hasNativeReflectConstruct = _isNativeReflectConstruct();

  return function _createSuperInternal() {
    var Super = _getPrototypeOf(Derived),
        result;

    if (hasNativeReflectConstruct) {
      var NewTarget = _getPrototypeOf(this).constructor;

      result = Reflect.construct(Super, arguments, NewTarget);
    } else {
      result = Super.apply(this, arguments);
    }

    return _possibleConstructorReturn(this, result);
  };
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _i = arr && (typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"]);

  if (_i == null) return;
  var _arr = [];
  var _n = true;
  var _d = false;

  var _s, _e;

  try {
    for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __decorate(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}

function __metadata(metadataKey, metadataValue) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(metadataKey, metadataValue);
}

var toString = {}.toString;
var isType = function (value, type) { return toString.call(value) === '[object ' + type + ']'; };

// isFinite,
var isNil = function (value) {
    /**
     * isNil(null) => true
     * isNil() => true
     */
    return value === null || value === undefined;
};

var isArray = (function (value) {
    return Array.isArray ?
        Array.isArray(value) :
        isType(value, 'Array');
});

var isString = (function (str) {
    return isType(str, 'String');
});

var mod = function (n, m) {
    return ((n % m) + m) % m;
};

var RADIAN = Math.PI / 180;
var toRadian = function (degree) {
    return RADIAN * degree;
};

var DefaultRenderer = /*#__PURE__*/function () {
  function DefaultRenderer() {
    _classCallCheck(this, DefaultRenderer);
  }

  _createClass(DefaultRenderer, [{
    key: "render",
    value: function render(context, attributes) {
      var fill = attributes.fill,
          opacity = attributes.opacity,
          fillOpacity = attributes.fillOpacity,
          stroke = attributes.stroke,
          strokeOpacity = attributes.strokeOpacity,
          lineWidth = attributes.lineWidth,
          lineCap = attributes.lineCap,
          lineJoin = attributes.lineJoin;

      if (!isNil(fill)) {
        if (!isNil(fillOpacity) && fillOpacity !== 1) {
          context.globalAlpha = fillOpacity;
          context.fill();
          context.globalAlpha = opacity;
        } else {
          context.fill();
        }
      }

      if (!isNil(stroke)) {
        if (lineWidth && lineWidth > 0) {
          var applyOpacity = !isNil(strokeOpacity) && strokeOpacity !== 1;

          if (applyOpacity) {
            context.globalAlpha = strokeOpacity;
          }

          context.lineWidth = lineWidth;

          if (!isNil(lineCap)) {
            context.lineCap = lineCap;
          }

          if (!isNil(lineJoin)) {
            context.lineJoin = lineJoin;
          }

          context.stroke();
        }
      }
    }
  }]);

  return DefaultRenderer;
}();

DefaultRenderer = __decorate([injectable()], DefaultRenderer);

var ImagePool = /*#__PURE__*/function () {
  function ImagePool() {
    _classCallCheck(this, ImagePool);

    this.imageCache = {};
  }

  _createClass(ImagePool, [{
    key: "getImageSync",
    value: function getImageSync(src) {
      return this.imageCache[src];
    }
  }, {
    key: "getOrCreateImage",
    value: function getOrCreateImage(src) {
      var _this = this;

      if (this.imageCache[src]) {
        return Promise.resolve(this.imageCache[src]);
      }

      return new Promise(function (resolve, reject) {
        var image = new Image();

        image.onload = function () {
          resolve(image);
        };

        image.onerror = function (ev) {
          reject(ev);
        };

        image.crossOrigin = 'Anonymous';
        image.src = src;
        _this.imageCache[src] = image;
      });
    }
  }]);

  return ImagePool;
}();

ImagePool = __decorate([injectable()], ImagePool);

var ImageRenderer = /*#__PURE__*/function () {
  function ImageRenderer() {
    _classCallCheck(this, ImageRenderer);

    this.imagePool = void 0;
  }

  _createClass(ImageRenderer, [{
    key: "render",
    value: function render(context, attributes) {
      // @ts-ignore
      var _attributes$width = attributes.width,
          width = _attributes$width === void 0 ? 0 : _attributes$width,
          _attributes$height = attributes.height,
          height = _attributes$height === void 0 ? 0 : _attributes$height,
          img = attributes.img,
          sx = attributes.sx,
          sy = attributes.sy,
          swidth = attributes.swidth,
          sheight = attributes.sheight,
          _attributes$anchor = attributes.anchor,
          anchor = _attributes$anchor === void 0 ? [0, 0] : _attributes$anchor;
      var image;
      var iw = width;
      var ih = height;

      if (isString(img)) {
        // image has been loaded in `mounted` hook
        image = this.imagePool.getImageSync(img);
      } else {
        iw || (iw = img.width);
        ih || (ih = img.height);
        image = img;
      }

      if (!isNil(sx) && !isNil(sy) && !isNil(swidth) && !isNil(sheight)) {
        context.drawImage(image, sx, sy, swidth, sheight, -anchor[0] * iw, -anchor[1] * ih, iw, ih);
      } else {
        context.drawImage(image, -anchor[0] * iw, -anchor[1] * ih, iw, ih);
      }
    }
  }]);

  return ImageRenderer;
}();

__decorate([inject(ImagePool), __metadata("design:type", ImagePool)], ImageRenderer.prototype, "imagePool", void 0);

ImageRenderer = __decorate([injectable()], ImageRenderer);

var TextRenderer = /*#__PURE__*/function () {
  function TextRenderer() {
    _classCallCheck(this, TextRenderer);

    this.textService = void 0;
  }

  _createClass(TextRenderer, [{
    key: "render",
    value: function render(context, attributes) {
      var _attributes$text = attributes.text,
          text = _attributes$text === void 0 ? '' : _attributes$text,
          _attributes$lineWidth = attributes.lineWidth,
          lineWidth = _attributes$lineWidth === void 0 ? 0 : _attributes$lineWidth,
          textAlign = attributes.textAlign,
          textBaseline = attributes.textBaseline,
          lineJoin = attributes.lineJoin,
          _attributes$miterLimi = attributes.miterLimit,
          miterLimit = _attributes$miterLimi === void 0 ? 0 : _attributes$miterLimi,
          _attributes$padding = attributes.padding,
          padding = _attributes$padding === void 0 ? 0 : _attributes$padding,
          _attributes$letterSpa = attributes.letterSpacing,
          letterSpacing = _attributes$letterSpa === void 0 ? 0 : _attributes$letterSpa,
          stroke = attributes.stroke,
          fill = attributes.fill,
          fillOpacity = attributes.fillOpacity,
          strokeOpacity = attributes.strokeOpacity,
          opacity = attributes.opacity;

      var _this$textService$mea = this.textService.measureText(text, attributes),
          font = _this$textService$mea.font,
          lines = _this$textService$mea.lines,
          height = _this$textService$mea.height,
          lineHeight = _this$textService$mea.lineHeight;

      context.font = font;
      context.lineWidth = lineWidth;
      context.textBaseline = textBaseline;
      context.lineJoin = lineJoin;
      context.miterLimit = miterLimit;
      var linePositionY = 0; // handle vertical text baseline

      if (textBaseline === 'middle') {
        linePositionY = -height / 2 - lineHeight / 2;
      } else if (textBaseline === 'bottom' || textBaseline === 'alphabetic' || textBaseline === 'ideographic') {
        linePositionY = -height;
      } else if (textBaseline === 'top' || textBaseline === 'hanging') {
        linePositionY = -lineHeight;
      } // draw lines line by line


      for (var i = 0; i < lines.length; i++) {
        var linePositionX = lineWidth / 2;
        linePositionY += lineHeight; // no need to re-position X, cause we already set text align
        // @see https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/textAlign

        if (!isNil(stroke) && lineWidth) {
          this.drawLetterSpacing(context, lines[i], linePositionX + padding, linePositionY + padding, letterSpacing, fillOpacity, strokeOpacity, opacity, true);
        }

        if (!isNil(fill)) {
          this.drawLetterSpacing(context, lines[i], linePositionX + padding, linePositionY + padding, letterSpacing, fillOpacity, strokeOpacity, opacity);
        }
      }
    }
  }, {
    key: "drawLetterSpacing",
    value: function drawLetterSpacing(context, text, x, y, letterSpacing, fillOpacity, strokeOpacity, opacity) {
      var isStroke = arguments.length > 8 && arguments[8] !== undefined ? arguments[8] : false;

      // letterSpacing of 0 means normal
      if (letterSpacing === 0) {
        if (isStroke) {
          this.strokeText(context, text, x, y, strokeOpacity);
        } else {
          this.fillText(context, text, x, y, fillOpacity, opacity);
        }

        return;
      }

      var currentPosition = x;
      var stringArray = Array.from(text);
      var previousWidth = context.measureText(text).width;
      var currentWidth = 0;

      for (var i = 0; i < stringArray.length; ++i) {
        var currentChar = stringArray[i];

        if (isStroke) {
          this.strokeText(context, currentChar, currentPosition, y, strokeOpacity);
        } else {
          this.fillText(context, currentChar, currentPosition, y, fillOpacity, opacity);
        }

        currentWidth = context.measureText(text.substring(i + 1)).width;
        currentPosition += previousWidth - currentWidth + letterSpacing;
        previousWidth = currentWidth;
      }
    }
  }, {
    key: "fillText",
    value: function fillText(context, text, x, y, fillOpacity, opacity) {
      var applyOpacity = !isNil(fillOpacity) && fillOpacity !== 1;

      if (applyOpacity) {
        context.globalAlpha = fillOpacity;
      }

      context.fillText(text, x, y);

      if (applyOpacity) {
        context.globalAlpha = opacity;
      }
    }
  }, {
    key: "strokeText",
    value: function strokeText(context, text, x, y, strokeOpacity) {
      var applyOpacity = !isNil(strokeOpacity) && strokeOpacity !== 1;

      if (applyOpacity) {
        context.globalAlpha = strokeOpacity;
      }

      context.strokeText(text, x, y);
    }
  }]);

  return TextRenderer;
}();

__decorate([inject(TextService), __metadata("design:type", TextService)], TextRenderer.prototype, "textService", void 0);

TextRenderer = __decorate([injectable()], TextRenderer);

var StyleRendererFactory = Symbol('StyleRendererFactory');

var StyleParser = /*#__PURE__*/function () {
  function StyleParser() {
    _classCallCheck(this, StyleParser);
  }

  _createClass(StyleParser, [{
    key: "parse",
    value: // @inject(ContextService)
    // protected contextService: ContextService<CanvasRenderingContext2D>;
    function parse(color) {
      // if (color[1] === '(' || color[2] === '(') {
      //   if (color[0] === 'l') {
      //     // regexLG.test(color)
      //     return parseLineGradient(context, element, color);
      //   }
      //   if (color[0] === 'r') {
      //     // regexRG.test(color)
      //     return parseRadialGradient(context, element, color);
      //   }
      //   if (color[0] === 'p') {
      //     // regexPR.test(color)
      //     return parsePattern(context, element, color);
      //   }
      // }
      return color;
    }
  }]);

  return StyleParser;
}();

StyleParser = __decorate([injectable()], StyleParser);

function generatePath(context, attributes) {
  context.beginPath();
  var _attributes$r = attributes.r,
      r = _attributes$r === void 0 ? 0 : _attributes$r;
  context.arc(0, 0, r, 0, Math.PI * 2, false);
}

function generatePath$1(context, attributes) {
  var rx = attributes.rx,
      ry = attributes.ry;
  context.beginPath(); // @see https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/ellipse

  if (context.ellipse) {
    context.ellipse(0, 0, rx, ry, 0, 0, Math.PI * 2, false);
  } else {
    // 如果不支持，则使用圆来绘制，进行变形
    var r = rx > ry ? rx : ry;
    var scaleX = rx > ry ? 1 : rx / ry;
    var scaleY = rx > ry ? ry / rx : 1;
    context.save();
    context.scale(scaleX, scaleY);
    context.arc(0, 0, r, 0, Math.PI * 2);
    context.closePath();
    context.restore();
  }
}

// const regexRG = /^r\s*\(\s*([\d.]+)\s*,\s*([\d.]+)\s*,\s*([\d.]+)\s*\)\s*(.*)/i;
// const regexPR = /^p\s*\(\s*([axyn])\s*\)\s*(.*)/i;
// const regexColorStop = /[\d.]+:(#[^\s]+|[^\)]+\))/gi;
// function addStop(steps, gradient: CanvasGradient) {
//   const arr: string[] = steps.match(regexColorStop);
//   each(arr, (item) => {
//     const itemArr = item.split(':');
//     gradient.addColorStop(itemArr[0], itemArr[1]);
//   });
// }
// export function parseLineGradient(context: CanvasRenderingContext2D, element: IElement, gradientStr: string) {
//   const arr = regexLG.exec(gradientStr);
//   const angle = (parseFloat(arr[1]) % 360) * (Math.PI / 180);
//   const steps = arr[2];
//   const box = element.getBBox();
//   let start;
//   let end;
//   if (angle >= 0 && angle < (1 / 2) * Math.PI) {
//     start = {
//       x: box.minX,
//       y: box.minY,
//     };
//     end = {
//       x: box.maxX,
//       y: box.maxY,
//     };
//   } else if ((1 / 2) * Math.PI <= angle && angle < Math.PI) {
//     start = {
//       x: box.maxX,
//       y: box.minY,
//     };
//     end = {
//       x: box.minX,
//       y: box.maxY,
//     };
//   } else if (Math.PI <= angle && angle < (3 / 2) * Math.PI) {
//     start = {
//       x: box.maxX,
//       y: box.maxY,
//     };
//     end = {
//       x: box.minX,
//       y: box.minY,
//     };
//   } else {
//     start = {
//       x: box.minX,
//       y: box.maxY,
//     };
//     end = {
//       x: box.maxX,
//       y: box.minY,
//     };
//   }
//   const tanTheta = Math.tan(angle);
//   const tanTheta2 = tanTheta * tanTheta;
//   const x = (end.x - start.x + tanTheta * (end.y - start.y)) / (tanTheta2 + 1) + start.x;
//   const y = (tanTheta * (end.x - start.x + tanTheta * (end.y - start.y))) / (tanTheta2 + 1) + start.y;
//   const gradient = context.createLinearGradient(start.x, start.y, x, y);
//   addStop(steps, gradient);
//   return gradient;
// }
// export function parseRadialGradient(context: CanvasRenderingContext2D, element: IElement, gradientStr: string) {
//   const arr = regexRG.exec(gradientStr);
//   const fx = parseFloat(arr[1]);
//   const fy = parseFloat(arr[2]);
//   const fr = parseFloat(arr[3]);
//   const steps = arr[4];
//   // 环半径为0时，默认无渐变，取渐变序列的最后一个颜色
//   if (fr === 0) {
//     const colors = steps.match(regexColorStop);
//     return colors[colors.length - 1].split(':')[1];
//   }
//   const box = element.getBBox();
//   const width = box.maxX - box.minX;
//   const height = box.maxY - box.minY;
//   const r = Math.sqrt(width * width + height * height) / 2;
//   const gradient = context.createRadialGradient(
//     box.minX + width * fx,
//     box.minY + height * fy,
//     0,
//     box.minX + width / 2,
//     box.minY + height / 2,
//     fr * r
//   );
//   addStop(steps, gradient);
//   return gradient;
// }
// /**
//  * 边和填充设置的颜色转换成 pattern
//  * @param {CanvasRenderingContext2D} context canvas 上下文
//  * @param {IElement}                 element  图形元素
//  * @param {string}                   patternStr   生成 pattern 的字符串
//  */
// export function parsePattern(context: CanvasRenderingContext2D, element: IElement, patternStr: string) {
//   // 在转换过程中进行了缓存
//   if (element.get('patternSource') && element.get('patternSource') === patternStr) {
//     return element.get('pattern');
//   }
//   let pattern;
//   let img: HTMLImageElement;
//   const arr = regexPR.exec(patternStr);
//   let repeat = arr[1];
//   const source = arr[2];
//   // Function to be called when pattern loads
//   function onload() {
//     // Create pattern
//     pattern = context.createPattern(img, repeat);
//     element.set('pattern', pattern); // be a cache
//     element.set('patternSource', patternStr);
//   }
//   switch (repeat) {
//     case 'a':
//       repeat = 'repeat';
//       break;
//     case 'x':
//       repeat = 'repeat-x';
//       break;
//     case 'y':
//       repeat = 'repeat-y';
//       break;
//     case 'n':
//       repeat = 'no-repeat';
//       break;
//     default:
//       repeat = 'no-repeat';
//   }
//   img = new Image();
//   // If source URL is not a data URL
//   if (!source.match(/^data:/i)) {
//     // Set crossOrigin for this image
//     img.crossOrigin = 'Anonymous';
//   }
//   img.src = source;
//   if (img.complete) {
//     onload();
//   } else {
//     img.onload = onload;
//     // Fix onload() bug in IE9
//     img.src = img.src;
//   }
//   return pattern;
// }
// export function parseStyle(context: CanvasRenderingContext2D, element: IElement, color: string) {
//   if (isString(color)) {
//     if (color[1] === '(' || color[2] === '(') {
//       if (color[0] === 'l') {
//         // regexLG.test(color)
//         return parseLineGradient(context, element, color);
//       }
//       if (color[0] === 'r') {
//         // regexRG.test(color)
//         return parseRadialGradient(context, element, color);
//       }
//       if (color[0] === 'p') {
//         // regexPR.test(color)
//         return parsePattern(context, element, color);
//       }
//     }
//     return color;
//   }
// }

function parseRadius(radius) {
  var r1 = 0;
  var r2 = 0;
  var r3 = 0;
  var r4 = 0;

  if (isArray(radius)) {
    if (radius.length === 1) {
      r1 = r2 = r3 = r4 = radius[0];
    } else if (radius.length === 2) {
      r1 = r3 = radius[0];
      r2 = r4 = radius[1];
    } else if (radius.length === 3) {
      r1 = radius[0];
      r2 = r4 = radius[1];
      r3 = radius[2];
    } else {
      r1 = radius[0];
      r2 = radius[1];
      r3 = radius[2];
      r4 = radius[3];
    }
  } else {
    r1 = r2 = r3 = r4 = radius;
  }

  return [r1, r2, r3, r4];
}

function generatePath$2(context, attributes) {
  var _attributes$radius = attributes.radius,
      radius = _attributes$radius === void 0 ? 0 : _attributes$radius,
      _attributes$width = attributes.width,
      width = _attributes$width === void 0 ? 0 : _attributes$width,
      _attributes$height = attributes.height,
      height = _attributes$height === void 0 ? 0 : _attributes$height;
  context.beginPath();

  if (radius === 0) {
    context.rect(0, 0, width, height);
  } else {
    var _parseRadius = parseRadius(radius),
        _parseRadius2 = _slicedToArray(_parseRadius, 4),
        r1 = _parseRadius2[0],
        r2 = _parseRadius2[1],
        r3 = _parseRadius2[2],
        r4 = _parseRadius2[3];

    context.moveTo(r1, 0);
    context.lineTo(width - r2, 0);
    r2 !== 0 && context.arc(width - r2, r2, r2, -Math.PI / 2, 0);
    context.lineTo(width, height - r3);
    r3 !== 0 && context.arc(width - r3, height - r3, r3, 0, Math.PI / 2);
    context.lineTo(r4, height);
    r4 !== 0 && context.arc(r4, height - r4, r4, Math.PI / 2, Math.PI);
    context.lineTo(0, r1);
    r1 !== 0 && context.arc(r1, r1, r1, Math.PI, Math.PI * 1.5);
    context.closePath();
  }
}

function generatePath$3(context, attributes) {
  var x1 = attributes.x1,
      y1 = attributes.y1,
      x2 = attributes.x2,
      y2 = attributes.y2,
      _attributes$x = attributes.x,
      x = _attributes$x === void 0 ? 0 : _attributes$x,
      _attributes$y = attributes.y,
      y = _attributes$y === void 0 ? 0 : _attributes$y;
  context.beginPath();
  context.moveTo(x1 - x, y1 - y);
  context.lineTo(x2 - x, y2 - y);
}

function generatePath$4(context, attributes) {
  var _attributes$x = attributes.x,
      x = _attributes$x === void 0 ? 0 : _attributes$x,
      _attributes$y = attributes.y,
      y = _attributes$y === void 0 ? 0 : _attributes$y; // @ts-ignore

  var points = attributes.points;
  var length = points.length;
  var x1 = points[0][0] - x;
  var y1 = points[0][1] - y;
  var x2 = points[length - 1][0] - x;
  var y2 = points[length - 1][1] - y;
  context.beginPath();
  context.moveTo(x1, y1);

  for (var i = 0; i < length - 1; i++) {
    var point = points[i];
    context.lineTo(point[0] - x, point[1] - y);
  }

  context.lineTo(x2, y2);
}

function generatePath$5(context, attributes) {
  var _attributes$x = attributes.x,
      x = _attributes$x === void 0 ? 0 : _attributes$x,
      _attributes$y = attributes.y,
      y = _attributes$y === void 0 ? 0 : _attributes$y; // @ts-ignore

  var points = attributes.points;
  var length = points.length;
  var x1 = points[0][0] - x;
  var y1 = points[0][1] - y;
  var x2 = points[length - 1][0] - x;
  var y2 = points[length - 1][1] - y;
  context.beginPath();
  context.moveTo(x1, y1);

  for (var i = 0; i < length - 1; i++) {
    var point = points[i];
    context.lineTo(point[0] - x, point[1] - y);
  }

  context.lineTo(x2, y2);
  context.closePath();
}

/**
 * 判断两个点是否重合，点坐标的格式为 [x, y]
 */

function isSamePoint(point1, point2) {
  return point1[0] === point2[0] && point1[1] === point2[1];
}

function vMag(v) {
  return Math.sqrt(v[0] * v[0] + v[1] * v[1]);
} // u.v/|u||v|，计算夹角的余弦值


function vRatio(u, v) {
  // 当存在一个向量的长度为 0 时，夹角也为 0，即夹角的余弦值为 1
  return vMag(u) * vMag(v) ? (u[0] * v[0] + u[1] * v[1]) / (vMag(u) * vMag(v)) : 1;
} // 向量角度


function vAngle(u, v) {
  return (u[0] * v[1] < u[1] * v[0] ? -1 : 1) * Math.acos(vRatio(u, v));
} // A 0:rx 1:ry 2:x-axis-rotation 3:large-arc-flag 4:sweep-flag 5: x 6: y


function getArcParams(startPoint, params, x, y) {
  var rx = params[1];
  var ry = params[2];
  var xRotation = mod(toRadian(params[3]), Math.PI * 2);
  var arcFlag = params[4];
  var sweepFlag = params[5]; // 弧形起点坐标

  var x1 = startPoint[0];
  var y1 = startPoint[1]; // 弧形终点坐标

  var x2 = params[6] - x;
  var y2 = params[7] - y;
  var xp = Math.cos(xRotation) * (x1 - x2) / 2.0 + Math.sin(xRotation) * (y1 - y2) / 2.0;
  var yp = -1 * Math.sin(xRotation) * (x1 - x2) / 2.0 + Math.cos(xRotation) * (y1 - y2) / 2.0;
  var lambda = xp * xp / (rx * rx) + yp * yp / (ry * ry);

  if (lambda > 1) {
    rx *= Math.sqrt(lambda);
    ry *= Math.sqrt(lambda);
  }

  var diff = rx * rx * (yp * yp) + ry * ry * (xp * xp);
  var f = diff ? Math.sqrt((rx * rx * (ry * ry) - diff) / diff) : 1;

  if (arcFlag === sweepFlag) {
    f *= -1;
  }

  if (isNaN(f)) {
    f = 0;
  } // 旋转前的起点坐标，且当长半轴和短半轴的长度为 0 时，坐标按 (0, 0) 处理


  var cxp = ry ? f * rx * yp / ry : 0;
  var cyp = rx ? f * -ry * xp / rx : 0; // 椭圆圆心坐标

  var cx = (x1 + x2) / 2.0 + Math.cos(xRotation) * cxp - Math.sin(xRotation) * cyp;
  var cy = (y1 + y2) / 2.0 + Math.sin(xRotation) * cxp + Math.cos(xRotation) * cyp; // 起始点的单位向量

  var u = [(xp - cxp) / rx, (yp - cyp) / ry]; // 终止点的单位向量

  var v = [(-1 * xp - cxp) / rx, (-1 * yp - cyp) / ry]; // 计算起始点和圆心的连线，与 x 轴正方向的夹角

  var theta = vAngle([1, 0], u); // 计算圆弧起始点和终止点与椭圆圆心连线的夹角

  var dTheta = vAngle(u, v);

  if (vRatio(u, v) <= -1) {
    dTheta = Math.PI;
  }

  if (vRatio(u, v) >= 1) {
    dTheta = 0;
  }

  if (sweepFlag === 0 && dTheta > 0) {
    dTheta = dTheta - 2 * Math.PI;
  }

  if (sweepFlag === 1 && dTheta < 0) {
    dTheta = dTheta + 2 * Math.PI;
  }

  return {
    cx: cx,
    cy: cy,
    // 弧形的起点和终点相同时，长轴和短轴的长度按 0 处理
    rx: isSamePoint(startPoint, [x2, y2]) ? 0 : rx,
    ry: isSamePoint(startPoint, [x2, y2]) ? 0 : ry,
    startAngle: theta,
    endAngle: theta + dTheta,
    xRotation: xRotation,
    arcFlag: arcFlag,
    sweepFlag: sweepFlag
  };
}

function generatePath$6(context, attributes) {
  // @ts-ignore
  var _attributes$x = attributes.x,
      x = _attributes$x === void 0 ? 0 : _attributes$x,
      _attributes$y = attributes.y,
      y = _attributes$y === void 0 ? 0 : _attributes$y,
      path = attributes.path,
      arcParamsCache = attributes.arcParamsCache;

  if (!arcParamsCache) {
    // @ts-ignore
    attributes.arcParamsCache = {};
  }

  var currentPoint = [-x, -y]; // 当前图形

  var startMovePoint = [-x, -y]; // 开始 M 的点，可能会有多个

  context.beginPath();

  for (var i = 0; i < path.length; i++) {
    var params = path[i]; // eg. M 100 200

    var command = params[0]; // V,H,S,T 都在前面被转换成标准形式

    switch (command) {
      case 'M':
        context.moveTo(params[1] - x, params[2] - y);
        startMovePoint = [params[1] - x, params[2] - y];
        break;

      case 'L':
        context.lineTo(params[1] - x, params[2] - y);
        break;

      case 'Q':
        context.quadraticCurveTo(params[1] - x, params[2] - y, params[3] - x, params[4] - y);
        break;

      case 'C':
        context.bezierCurveTo(params[1] - x, params[2] - y, params[3] - x, params[4] - y, params[5] - x, params[6] - y);
        break;

      case 'A':
        {
          var arcParams = void 0; // 为了加速绘制，可以提供参数的缓存，各个图形自己缓存

          if (arcParamsCache) {
            arcParams = arcParamsCache[i];

            if (!arcParams) {
              arcParams = getArcParams(currentPoint, params, x, y);
              arcParamsCache[i] = arcParams;
            }
          } else {
            arcParams = getArcParams(currentPoint, params, x, y);
          }

          var _arcParams = arcParams,
              cx = _arcParams.cx,
              cy = _arcParams.cy,
              rx = _arcParams.rx,
              ry = _arcParams.ry,
              startAngle = _arcParams.startAngle,
              endAngle = _arcParams.endAngle,
              xRotation = _arcParams.xRotation,
              sweepFlag = _arcParams.sweepFlag; // 直接使用椭圆的 api

          if (context.ellipse) {
            context.ellipse(cx, cy, rx, ry, xRotation, startAngle, endAngle, !!(1 - sweepFlag));
          } else {
            // 如果不支持，则使用圆来绘制，进行变形
            var r = rx > ry ? rx : ry;
            var scaleX = rx > ry ? 1 : rx / ry;
            var scaleY = rx > ry ? ry / rx : 1;
            context.translate(cx, cy);
            context.rotate(xRotation);
            context.scale(scaleX, scaleY);
            context.arc(0, 0, r, startAngle, endAngle, !!(1 - sweepFlag));
            context.scale(1 / scaleX, 1 / scaleY);
            context.rotate(-xRotation);
            context.translate(-cx, -cy);
          }

          break;
        }

      case 'Z':
        context.closePath();
        break;
    } // 有了 Z 后，当前节点从开始 M 的点开始


    if (command === 'Z') {
      currentPoint = startMovePoint;
    } else {
      var len = params.length;
      currentPoint = [params[len - 2] - x, params[len - 1] - y];
    }
  }
}

var PathGeneratorFactory = Symbol('PathGeneratorFactory');
var PathGenerator = Symbol('Path');

/**
 * Common utilities
 * @module glMatrix
 */
// Configuration Constants
var EPSILON = 0.000001;
var ARRAY_TYPE = typeof Float32Array !== 'undefined' ? Float32Array : Array;
if (!Math.hypot) Math.hypot = function () {
  var y = 0,
      i = arguments.length;

  while (i--) {
    y += arguments[i] * arguments[i];
  }

  return Math.sqrt(y);
};

/**
 * 3x3 Matrix
 * @module mat3
 */

/**
 * Creates a new identity mat3
 *
 * @returns {mat3} a new 3x3 matrix
 */

function create() {
  var out = new ARRAY_TYPE(9);

  if (ARRAY_TYPE != Float32Array) {
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[5] = 0;
    out[6] = 0;
    out[7] = 0;
  }

  out[0] = 1;
  out[4] = 1;
  out[8] = 1;
  return out;
}

/**
 * Returns the translation vector component of a transformation
 *  matrix. If a matrix is built with fromRotationTranslation,
 *  the returned vector will be the same as the translation vector
 *  originally supplied.
 * @param  {vec3} out Vector to receive translation component
 * @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {vec3} out
 */

function getTranslation(out, mat) {
  out[0] = mat[12];
  out[1] = mat[13];
  out[2] = mat[14];
  return out;
}
/**
 * Returns the scaling factor component of a transformation
 *  matrix. If a matrix is built with fromRotationTranslationScale
 *  with a normalized Quaternion paramter, the returned vector will be
 *  the same as the scaling vector
 *  originally supplied.
 * @param  {vec3} out Vector to receive scaling factor component
 * @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {vec3} out
 */

function getScaling(out, mat) {
  var m11 = mat[0];
  var m12 = mat[1];
  var m13 = mat[2];
  var m21 = mat[4];
  var m22 = mat[5];
  var m23 = mat[6];
  var m31 = mat[8];
  var m32 = mat[9];
  var m33 = mat[10];
  out[0] = Math.hypot(m11, m12, m13);
  out[1] = Math.hypot(m21, m22, m23);
  out[2] = Math.hypot(m31, m32, m33);
  return out;
}
/**
 * Returns a quaternion representing the rotational component
 *  of a transformation matrix. If a matrix is built with
 *  fromRotationTranslation, the returned quaternion will be the
 *  same as the quaternion originally supplied.
 * @param {quat} out Quaternion to receive the rotation component
 * @param {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {quat} out
 */

function getRotation(out, mat) {
  var scaling = new ARRAY_TYPE(3);
  getScaling(scaling, mat);
  var is1 = 1 / scaling[0];
  var is2 = 1 / scaling[1];
  var is3 = 1 / scaling[2];
  var sm11 = mat[0] * is1;
  var sm12 = mat[1] * is2;
  var sm13 = mat[2] * is3;
  var sm21 = mat[4] * is1;
  var sm22 = mat[5] * is2;
  var sm23 = mat[6] * is3;
  var sm31 = mat[8] * is1;
  var sm32 = mat[9] * is2;
  var sm33 = mat[10] * is3;
  var trace = sm11 + sm22 + sm33;
  var S = 0;

  if (trace > 0) {
    S = Math.sqrt(trace + 1.0) * 2;
    out[3] = 0.25 * S;
    out[0] = (sm23 - sm32) / S;
    out[1] = (sm31 - sm13) / S;
    out[2] = (sm12 - sm21) / S;
  } else if (sm11 > sm22 && sm11 > sm33) {
    S = Math.sqrt(1.0 + sm11 - sm22 - sm33) * 2;
    out[3] = (sm23 - sm32) / S;
    out[0] = 0.25 * S;
    out[1] = (sm12 + sm21) / S;
    out[2] = (sm31 + sm13) / S;
  } else if (sm22 > sm33) {
    S = Math.sqrt(1.0 + sm22 - sm11 - sm33) * 2;
    out[3] = (sm31 - sm13) / S;
    out[0] = (sm12 + sm21) / S;
    out[1] = 0.25 * S;
    out[2] = (sm23 + sm32) / S;
  } else {
    S = Math.sqrt(1.0 + sm33 - sm11 - sm22) * 2;
    out[3] = (sm12 - sm21) / S;
    out[0] = (sm31 + sm13) / S;
    out[1] = (sm23 + sm32) / S;
    out[2] = 0.25 * S;
  }

  return out;
}

/**
 * 3 Dimensional Vector
 * @module vec3
 */

/**
 * Creates a new, empty vec3
 *
 * @returns {vec3} a new 3D vector
 */

function create$1() {
  var out = new ARRAY_TYPE(3);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  return out;
}
/**
 * Calculates the length of a vec3
 *
 * @param {ReadonlyVec3} a vector to calculate length of
 * @returns {Number} length of a
 */

function length(a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  return Math.hypot(x, y, z);
}
/**
 * Creates a new vec3 initialized with the given values
 *
 * @param {Number} x X component
 * @param {Number} y Y component
 * @param {Number} z Z component
 * @returns {vec3} a new 3D vector
 */

function fromValues(x, y, z) {
  var out = new ARRAY_TYPE(3);
  out[0] = x;
  out[1] = y;
  out[2] = z;
  return out;
}
/**
 * Normalize a vec3
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a vector to normalize
 * @returns {vec3} out
 */

function normalize(out, a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  var len = x * x + y * y + z * z;

  if (len > 0) {
    //TODO: evaluate use of glm_invsqrt here?
    len = 1 / Math.sqrt(len);
  }

  out[0] = a[0] * len;
  out[1] = a[1] * len;
  out[2] = a[2] * len;
  return out;
}
/**
 * Calculates the dot product of two vec3's
 *
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {Number} dot product of a and b
 */

function dot(a, b) {
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}
/**
 * Computes the cross product of two vec3's
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {vec3} out
 */

function cross(out, a, b) {
  var ax = a[0],
      ay = a[1],
      az = a[2];
  var bx = b[0],
      by = b[1],
      bz = b[2];
  out[0] = ay * bz - az * by;
  out[1] = az * bx - ax * bz;
  out[2] = ax * by - ay * bx;
  return out;
}
/**
 * Alias for {@link vec3.length}
 * @function
 */

var len = length;
/**
 * Perform some operation over an array of vec3s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec3. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec3s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach = function () {
  var vec = create$1();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 3;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
    }

    return a;
  };
}();

/**
 * 4 Dimensional Vector
 * @module vec4
 */

/**
 * Creates a new, empty vec4
 *
 * @returns {vec4} a new 4D vector
 */

function create$2() {
  var out = new ARRAY_TYPE(4);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
  }

  return out;
}
/**
 * Normalize a vec4
 *
 * @param {vec4} out the receiving vector
 * @param {ReadonlyVec4} a vector to normalize
 * @returns {vec4} out
 */

function normalize$1(out, a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  var w = a[3];
  var len = x * x + y * y + z * z + w * w;

  if (len > 0) {
    len = 1 / Math.sqrt(len);
  }

  out[0] = x * len;
  out[1] = y * len;
  out[2] = z * len;
  out[3] = w * len;
  return out;
}
/**
 * Perform some operation over an array of vec4s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec4. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec4s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach$1 = function () {
  var vec = create$2();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 4;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      vec[3] = a[i + 3];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
      a[i + 3] = vec[3];
    }

    return a;
  };
}();

/**
 * Quaternion
 * @module quat
 */

/**
 * Creates a new identity quat
 *
 * @returns {quat} a new quaternion
 */

function create$3() {
  var out = new ARRAY_TYPE(4);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  out[3] = 1;
  return out;
}
/**
 * Sets a quat from the given angle and rotation axis,
 * then returns it.
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyVec3} axis the axis around which to rotate
 * @param {Number} rad the angle in radians
 * @returns {quat} out
 **/

function setAxisAngle(out, axis, rad) {
  rad = rad * 0.5;
  var s = Math.sin(rad);
  out[0] = s * axis[0];
  out[1] = s * axis[1];
  out[2] = s * axis[2];
  out[3] = Math.cos(rad);
  return out;
}
/**
 * Performs a spherical linear interpolation between two quat
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a the first operand
 * @param {ReadonlyQuat} b the second operand
 * @param {Number} t interpolation amount, in the range [0-1], between the two inputs
 * @returns {quat} out
 */

function slerp(out, a, b, t) {
  // benchmarks:
  //    http://jsperf.com/quaternion-slerp-implementations
  var ax = a[0],
      ay = a[1],
      az = a[2],
      aw = a[3];
  var bx = b[0],
      by = b[1],
      bz = b[2],
      bw = b[3];
  var omega, cosom, sinom, scale0, scale1; // calc cosine

  cosom = ax * bx + ay * by + az * bz + aw * bw; // adjust signs (if necessary)

  if (cosom < 0.0) {
    cosom = -cosom;
    bx = -bx;
    by = -by;
    bz = -bz;
    bw = -bw;
  } // calculate coefficients


  if (1.0 - cosom > EPSILON) {
    // standard case (slerp)
    omega = Math.acos(cosom);
    sinom = Math.sin(omega);
    scale0 = Math.sin((1.0 - t) * omega) / sinom;
    scale1 = Math.sin(t * omega) / sinom;
  } else {
    // "from" and "to" quaternions are very close
    //  ... so we can do a linear interpolation
    scale0 = 1.0 - t;
    scale1 = t;
  } // calculate final values


  out[0] = scale0 * ax + scale1 * bx;
  out[1] = scale0 * ay + scale1 * by;
  out[2] = scale0 * az + scale1 * bz;
  out[3] = scale0 * aw + scale1 * bw;
  return out;
}
/**
 * Creates a quaternion from the given 3x3 rotation matrix.
 *
 * NOTE: The resultant quaternion is not normalized, so you should be sure
 * to renormalize the quaternion yourself where necessary.
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyMat3} m rotation matrix
 * @returns {quat} out
 * @function
 */

function fromMat3(out, m) {
  // Algorithm in Ken Shoemake's article in 1987 SIGGRAPH course notes
  // article "Quaternion Calculus and Fast Animation".
  var fTrace = m[0] + m[4] + m[8];
  var fRoot;

  if (fTrace > 0.0) {
    // |w| > 1/2, may as well choose w > 1/2
    fRoot = Math.sqrt(fTrace + 1.0); // 2w

    out[3] = 0.5 * fRoot;
    fRoot = 0.5 / fRoot; // 1/(4w)

    out[0] = (m[5] - m[7]) * fRoot;
    out[1] = (m[6] - m[2]) * fRoot;
    out[2] = (m[1] - m[3]) * fRoot;
  } else {
    // |w| <= 1/2
    var i = 0;
    if (m[4] > m[0]) i = 1;
    if (m[8] > m[i * 3 + i]) i = 2;
    var j = (i + 1) % 3;
    var k = (i + 2) % 3;
    fRoot = Math.sqrt(m[i * 3 + i] - m[j * 3 + j] - m[k * 3 + k] + 1.0);
    out[i] = 0.5 * fRoot;
    fRoot = 0.5 / fRoot;
    out[3] = (m[j * 3 + k] - m[k * 3 + j]) * fRoot;
    out[j] = (m[j * 3 + i] + m[i * 3 + j]) * fRoot;
    out[k] = (m[k * 3 + i] + m[i * 3 + k]) * fRoot;
  }

  return out;
}
/**
 * Normalize a quat
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a quaternion to normalize
 * @returns {quat} out
 * @function
 */

var normalize$2 = normalize$1;
/**
 * Sets a quaternion to represent the shortest rotation from one
 * vector to another.
 *
 * Both vectors are assumed to be unit length.
 *
 * @param {quat} out the receiving quaternion.
 * @param {ReadonlyVec3} a the initial vector
 * @param {ReadonlyVec3} b the destination vector
 * @returns {quat} out
 */

var rotationTo = function () {
  var tmpvec3 = create$1();
  var xUnitVec3 = fromValues(1, 0, 0);
  var yUnitVec3 = fromValues(0, 1, 0);
  return function (out, a, b) {
    var dot$1 = dot(a, b);

    if (dot$1 < -0.999999) {
      cross(tmpvec3, xUnitVec3, a);
      if (len(tmpvec3) < 0.000001) cross(tmpvec3, yUnitVec3, a);
      normalize(tmpvec3, tmpvec3);
      setAxisAngle(out, tmpvec3, Math.PI);
      return out;
    } else if (dot$1 > 0.999999) {
      out[0] = 0;
      out[1] = 0;
      out[2] = 0;
      out[3] = 1;
      return out;
    } else {
      cross(tmpvec3, a, b);
      out[0] = tmpvec3[0];
      out[1] = tmpvec3[1];
      out[2] = tmpvec3[2];
      out[3] = 1 + dot$1;
      return normalize$2(out, out);
    }
  };
}();
/**
 * Performs a spherical linear interpolation with two control points
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a the first operand
 * @param {ReadonlyQuat} b the second operand
 * @param {ReadonlyQuat} c the third operand
 * @param {ReadonlyQuat} d the fourth operand
 * @param {Number} t interpolation amount, in the range [0-1], between the two inputs
 * @returns {quat} out
 */

var sqlerp = function () {
  var temp1 = create$3();
  var temp2 = create$3();
  return function (out, a, b, c, d, t) {
    slerp(temp1, a, d, t);
    slerp(temp2, b, c, t);
    slerp(out, temp1, temp2, 2 * t * (1 - t));
    return out;
  };
}();
/**
 * Sets the specified quaternion with values corresponding to the given
 * axes. Each axis is a vec3 and is expected to be unit length and
 * perpendicular to all other specified axes.
 *
 * @param {ReadonlyVec3} view  the vector representing the viewing direction
 * @param {ReadonlyVec3} right the vector representing the local "right" direction
 * @param {ReadonlyVec3} up    the vector representing the local "up" direction
 * @returns {quat} out
 */

var setAxes = function () {
  var matr = create();
  return function (out, view, right, up) {
    matr[0] = right[0];
    matr[3] = right[1];
    matr[6] = right[2];
    matr[1] = up[0];
    matr[4] = up[1];
    matr[7] = up[2];
    matr[2] = -view[0];
    matr[5] = -view[1];
    matr[8] = -view[2];
    return normalize$2(out, fromMat3(out, matr));
  };
}();

var RBushNode = /*#__PURE__*/function (_Component) {
  _inherits(RBushNode, _Component);

  var _super = _createSuper(RBushNode);

  function RBushNode() {
    var _this;

    _classCallCheck(this, RBushNode);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _super.call.apply(_super, [this].concat(args));
    _this.aabb = void 0;
    return _this;
  }

  return RBushNode;
}(Component);
RBushNode.tag = 'c-canvas-rbush-node';

var _class, _temp;

var CanvasRendererPlugin_1;
var RBushRoot = Symbol('RBushRoot');
var SHAPE_ATTRS_MAP = {
  fill: 'fillStyle',
  stroke: 'strokeStyle',
  opacity: 'globalAlpha'
};
/**
 * support 2 modes in rendering:
 * * immediate
 * * delayed: render at the end of frame with dirty-rectangle
 */

var CanvasRendererPlugin = CanvasRendererPlugin_1 = (_temp = _class = /*#__PURE__*/function () {
  function CanvasRendererPlugin() {
    var _this = this;

    _classCallCheck(this, CanvasRendererPlugin);

    this.canvasConfig = void 0;
    this.camera = void 0;
    this.contextService = void 0;
    this.sceneGraphService = void 0;
    this.renderingContext = void 0;
    this.styleParser = void 0;
    this.pathGeneratorFactory = void 0;
    this.styleRendererFactory = void 0;
    this.displayObjectPool = void 0;
    this.rBush = void 0;
    this.renderQueue = [];
    this.lastDirtyRectangle = void 0;

    this.handleEntityAABBChanged = function (object) {
      // skip if this object mounted on another scenegraph root
      if (object.ownerDocument !== _this.renderingContext.root) {
        return;
      }

      var entity = object.getEntity();

      var _this$canvasConfig$re = _this.canvasConfig.renderer.getConfig(),
          enableDirtyRectangleRendering = _this$canvasConfig$re.enableDirtyRectangleRendering; // don not use rbush when dirty-rectangle rendering disabled


      if (!enableDirtyRectangleRendering) {
        return;
      }

      var renderable = entity.getComponent(Renderable);
      var rBushNode = entity.getComponent(RBushNode);

      if (rBushNode) {
        // insert node in RTree
        if (rBushNode.aabb) {
          _this.rBush.remove(rBushNode.aabb);
        }

        if (renderable.aabb) {
          var _renderable$aabb$getM = renderable.aabb.getMin(),
              _renderable$aabb$getM2 = _slicedToArray(_renderable$aabb$getM, 2),
              minX = _renderable$aabb$getM2[0],
              minY = _renderable$aabb$getM2[1];

          var _renderable$aabb$getM3 = renderable.aabb.getMax(),
              _renderable$aabb$getM4 = _slicedToArray(_renderable$aabb$getM3, 2),
              maxX = _renderable$aabb$getM4[0],
              maxY = _renderable$aabb$getM4[1];

          rBushNode.aabb = {
            name: entity.getName(),
            minX: minX,
            minY: minY,
            maxX: maxX,
            maxY: maxY
          };
        }

        _this.rBush.insert(rBushNode.aabb);
      }
    };
  }

  _createClass(CanvasRendererPlugin, [{
    key: "apply",
    value: function apply(renderingService) {
      var _this2 = this;

      renderingService.hooks.init.tap(CanvasRendererPlugin_1.tag, function () {
        var context = _this2.contextService.getContext();

        var dpr = _this2.contextService.getDPR(); // scale all drawing operations by the dpr
        // @see https://www.html5rocks.com/en/tutorials/canvas/hidpi/


        context && context.scale(dpr, dpr);

        _this2.sceneGraphService.on(SCENE_GRAPH_EVENT.AABBChanged, _this2.handleEntityAABBChanged);
      });
      renderingService.hooks.mounted.tap(CanvasRendererPlugin_1.tag, function (object) {
        object.getEntity().addComponent(RBushNode);

        _this2.sceneGraphService.emit(SCENE_GRAPH_EVENT.AABBChanged, object);
      });
      renderingService.hooks.unmounted.tap(CanvasRendererPlugin_1.tag, function (object) {
        var rBushNode = object.getEntity().getComponent(RBushNode);

        _this2.rBush.remove(rBushNode.aabb);

        object.getEntity().removeComponent(RBushNode);
      });
      renderingService.hooks.destroy.tap(CanvasRendererPlugin_1.tag, function () {
        _this2.sceneGraphService.off(SCENE_GRAPH_EVENT.AABBChanged, _this2.handleEntityAABBChanged);
      });
      renderingService.hooks.beginFrame.tap(CanvasRendererPlugin_1.tag, function () {
        var context = _this2.contextService.getContext();

        var _this2$canvasConfig$r = _this2.canvasConfig.renderer.getConfig(),
            enableDirtyRectangleRendering = _this2$canvasConfig$r.enableDirtyRectangleRendering,
            enableDirtyRectangleRenderingDebug = _this2$canvasConfig$r.enableDirtyRectangleRenderingDebug;

        if (context) {
          context.save();

          if (!enableDirtyRectangleRendering) {
            context.clearRect(0, 0, _this2.canvasConfig.width, _this2.canvasConfig.height);
          } // account for camera's world matrix


          _this2.applyTransform(context, _this2.camera.getOrthoMatrix());
        }
      }); // render at the end of frame

      renderingService.hooks.endFrame.tap(CanvasRendererPlugin_1.tag, function () {
        var _this2$canvasConfig$r2 = _this2.canvasConfig.renderer.getConfig(),
            enableDirtyRectangleRendering = _this2$canvasConfig$r2.enableDirtyRectangleRendering;

        var context = _this2.contextService.getContext();

        if (enableDirtyRectangleRendering) {
          // merge removed AABB
          var dirtyAABB = _this2.safeMergeAABB.apply(_this2, [_this2.mergeDirtyAABBs(_this2.renderQueue)].concat(_toConsumableArray(_this2.renderingContext.removedAABBs)));

          _this2.renderingContext.removedAABBs = [];

          if (!dirtyAABB) {
            return;
          } // clear & clip dirty rectangle


          var _this2$convertAABB2Re = _this2.convertAABB2Rect(dirtyAABB),
              x = _this2$convertAABB2Re.x,
              y = _this2$convertAABB2Re.y,
              width = _this2$convertAABB2Re.width,
              height = _this2$convertAABB2Re.height;

          context.clearRect(x, y, width, height);
          context.beginPath();
          context.rect(x, y, width, height);
          context.clip(); // search objects intersect with dirty rectangle

          var dirtyObjects = _this2.searchDirtyObjects(dirtyAABB); // do rendering


          dirtyObjects // sort by z-index
          .sort(_this2.sceneGraphService.sort).forEach(function (object) {
            if (object.isVisible()) {
              _this2.renderDisplayObject(object);
            }
          }); // save dirty AABBs in last frame

          _this2.renderQueue.forEach(function (object) {
            _this2.saveDirtyAABB(object);
          }); // clear queue


          _this2.renderQueue = [];
        }

        context.restore();
      });
      renderingService.hooks.render.tap(CanvasRendererPlugin_1.tag, function (object) {
        var _this2$canvasConfig$r3 = _this2.canvasConfig.renderer.getConfig(),
            enableDirtyRectangleRendering = _this2$canvasConfig$r3.enableDirtyRectangleRendering;

        if (!enableDirtyRectangleRendering) {
          // render immediately
          _this2.renderDisplayObject(object);
        } else {
          // render at the end of frame
          _this2.renderQueue.push(object);
        }
      });
    }
  }, {
    key: "renderDisplayObject",
    value: function renderDisplayObject(object) {
      var context = this.contextService.getContext();
      var nodeName = object.nodeName; // reset transformation

      context.save(); // apply RTS transformation in world space

      this.applyTransform(context, object.getWorldTransform()); // apply attributes to context

      this.applyAttributesToContext(context, object.attributes); // clip path

      var clipPathShape = object.style.clipPath;

      if (clipPathShape) {
        context.save(); // apply clip shape's RTS

        this.applyTransform(context, clipPathShape.getWorldTransform());
        this.applyAttributesToContext(context, clipPathShape.attributes);

        var _generatePath = this.pathGeneratorFactory(clipPathShape.nodeName);

        if (_generatePath) {
          _generatePath(context, clipPathShape.attributes);
        }

        context.restore();
        context.clip();
      } // generate path in local space


      var generatePath = this.pathGeneratorFactory(nodeName);

      if (generatePath) {
        generatePath(context, object.attributes);
      } // fill & stroke


      var styleRenderer = this.styleRendererFactory(nodeName);

      if (styleRenderer) {
        styleRenderer.render(context, object.attributes);
      } // finish rendering, clear dirty flag


      var renderable = object.getEntity().getComponent(Renderable);
      renderable.dirty = false;
      context.restore();
    }
  }, {
    key: "convertAABB2Rect",
    value: function convertAABB2Rect(aabb) {
      var min = aabb.getMin();
      var max = aabb.getMax(); // expand the rectangle a bit to avoid artifacts
      // @see https://www.yuque.com/antv/ou292n/bi8nix#ExvCu

      var minX = Math.floor(min[0]);
      var minY = Math.floor(min[1]);
      var maxX = Math.ceil(max[0]);
      var maxY = Math.ceil(max[1]);
      var width = maxX - minX;
      var height = maxY - minY;
      return {
        x: minX,
        y: minY,
        width: width,
        height: height
      };
    }
    /**
     * TODO: merge dirty rectangles with some strategies.
     * For now, we just simply merge all the rectangles into one.
     * @see https://idom.me/articles/841.html
     */

  }, {
    key: "mergeDirtyAABBs",
    value: function mergeDirtyAABBs(dirtyObjects) {
      // merge into a big AABB
      var dirtyRectangle;
      dirtyObjects.forEach(function (object) {
        var aabb = object.getBounds();

        if (aabb) {
          if (!dirtyRectangle) {
            dirtyRectangle = new AABB(aabb.center, aabb.halfExtents);
          } else {
            dirtyRectangle.add(aabb);
          }
        }

        var _object$getEntity$get = object.getEntity().getComponent(Renderable),
            dirtyAABB = _object$getEntity$get.dirtyAABB;

        if (dirtyAABB) {
          if (!dirtyRectangle) {
            dirtyRectangle = new AABB(dirtyAABB.center, dirtyAABB.halfExtents);
          } else {
            dirtyRectangle.add(dirtyAABB);
          }
        }
      });
      return dirtyRectangle;
    }
  }, {
    key: "searchDirtyObjects",
    value: function searchDirtyObjects(dirtyRectangle) {
      var _this3 = this;

      // search in r-tree, get all affected nodes
      var _dirtyRectangle$getMi = dirtyRectangle.getMin(),
          _dirtyRectangle$getMi2 = _slicedToArray(_dirtyRectangle$getMi, 2),
          minX = _dirtyRectangle$getMi2[0],
          minY = _dirtyRectangle$getMi2[1];

      var _dirtyRectangle$getMa = dirtyRectangle.getMax(),
          _dirtyRectangle$getMa2 = _slicedToArray(_dirtyRectangle$getMa, 2),
          maxX = _dirtyRectangle$getMa2[0],
          maxY = _dirtyRectangle$getMa2[1];

      var rBushNodes = this.rBush.search({
        minX: minX,
        minY: minY,
        maxX: maxX,
        maxY: maxY
      });
      return rBushNodes.map(function (_ref) {
        var name = _ref.name;
        return _this3.displayObjectPool.getByName(name);
      });
    }
  }, {
    key: "saveDirtyAABB",
    value: function saveDirtyAABB(object) {
      var entity = object.getEntity();
      var renderable = entity.getComponent(Renderable);

      if (!renderable.dirtyAABB) {
        renderable.dirtyAABB = new AABB();
      }

      var bounds = object.getBounds();

      if (bounds) {
        // save last dirty aabb
        renderable.dirtyAABB.update(bounds.center, bounds.halfExtents);
      }
    }
  }, {
    key: "drawDirtyRectangle",
    value: function drawDirtyRectangle(context, _ref2) {
      var x = _ref2.x,
          y = _ref2.y,
          width = _ref2.width,
          height = _ref2.height;
      context.beginPath();
      context.rect(x + 1, y + 1, width - 1, height - 1);
      context.closePath();
      context.lineWidth = 1;
      context.stroke();
    }
  }, {
    key: "applyTransform",
    value: function applyTransform(context, transform) {
      var _mat4$getTranslation = getTranslation(create$1(), transform),
          _mat4$getTranslation2 = _slicedToArray(_mat4$getTranslation, 2),
          tx = _mat4$getTranslation2[0],
          ty = _mat4$getTranslation2[1];

      var _mat4$getScaling = getScaling(create$1(), transform),
          _mat4$getScaling2 = _slicedToArray(_mat4$getScaling, 2),
          sx = _mat4$getScaling2[0],
          sy = _mat4$getScaling2[1];

      var rotation = getRotation(create$3(), transform);

      var _getEuler = getEuler(create$1(), rotation),
          _getEuler2 = _slicedToArray(_getEuler, 3),
          eux = _getEuler2[0],
          euy = _getEuler2[1],
          euz = _getEuler2[2]; // gimbal lock at 90 degrees


      var rts = fromRotationTranslationScale(eux || euz, tx, ty, sx, sy); // @see https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Transformations

      context.transform(rts[0], rts[1], rts[3], rts[4], rts[6], rts[7]);
    }
  }, {
    key: "applyAttributesToContext",
    value: function applyAttributesToContext(context, attrs) {
      for (var k in attrs) {
        var v = attrs[k]; // 转换一下不与 canvas 兼容的属性名

        var name = SHAPE_ATTRS_MAP[k] ? SHAPE_ATTRS_MAP[k] : k;

        if (name === 'lineDash' && context.setLineDash) {
          isArray(v) && context.setLineDash(v);
        } else {
          if (name === 'strokeStyle' || name === 'fillStyle') {
            // 如果存在渐变、pattern 这个开销有些大
            // 可以考虑缓存机制，通过 hasUpdate 来避免一些运算
            v = this.styleParser.parse(v);
          } else if (name === 'globalAlpha') {
            // opacity 效果可以叠加，子元素的 opacity 需要与父元素 opacity 相乘
            v = v * context.globalAlpha;
          } // @ts-ignore


          context[name] = v;
        }
      }
    }
  }, {
    key: "safeMergeAABB",
    value: function safeMergeAABB() {
      var merged;

      for (var _len = arguments.length, aabbs = new Array(_len), _key = 0; _key < _len; _key++) {
        aabbs[_key] = arguments[_key];
      }

      aabbs.forEach(function (aabb) {
        if (aabb) {
          if (!merged) {
            merged = aabb;
          } else {
            merged.add(aabb);
          }
        }
      });
      return merged;
    }
  }]);

  return CanvasRendererPlugin;
}(), _class.tag = 'CanvasRendererPlugin', _temp);

__decorate([inject(CanvasConfig), __metadata("design:type", Object)], CanvasRendererPlugin.prototype, "canvasConfig", void 0);

__decorate([inject(Camera), __metadata("design:type", Camera)], CanvasRendererPlugin.prototype, "camera", void 0);

__decorate([inject(ContextService), __metadata("design:type", Object)], CanvasRendererPlugin.prototype, "contextService", void 0);

__decorate([inject(SceneGraphService), __metadata("design:type", SceneGraphService)], CanvasRendererPlugin.prototype, "sceneGraphService", void 0);

__decorate([inject(RenderingContext), __metadata("design:type", Object)], CanvasRendererPlugin.prototype, "renderingContext", void 0);

__decorate([inject(StyleParser), __metadata("design:type", StyleParser)], CanvasRendererPlugin.prototype, "styleParser", void 0);

__decorate([inject(PathGeneratorFactory), __metadata("design:type", Function)], CanvasRendererPlugin.prototype, "pathGeneratorFactory", void 0);

__decorate([inject(StyleRendererFactory), __metadata("design:type", Function)], CanvasRendererPlugin.prototype, "styleRendererFactory", void 0);

__decorate([inject(DisplayObjectPool), __metadata("design:type", DisplayObjectPool)], CanvasRendererPlugin.prototype, "displayObjectPool", void 0);

__decorate([inject(RBushRoot), __metadata("design:type", RBush)], CanvasRendererPlugin.prototype, "rBush", void 0);

CanvasRendererPlugin = CanvasRendererPlugin_1 = __decorate([injectable()], CanvasRendererPlugin);

var _class$1, _temp$1;

var LoadImagePlugin_1;
var LoadImagePlugin = LoadImagePlugin_1 = (_temp$1 = _class$1 = /*#__PURE__*/function () {
  function LoadImagePlugin() {
    _classCallCheck(this, LoadImagePlugin);

    this.imagePool = void 0;
  }

  _createClass(LoadImagePlugin, [{
    key: "apply",
    value: function apply(renderingService) {
      var _this = this;

      renderingService.hooks.mounted.tap(LoadImagePlugin_1.tag, function (object) {
        var nodeName = object.nodeName,
            attributes = object.attributes;

        if (nodeName === SHAPE.Image) {
          var img = attributes.img;

          if (isString(img)) {
            _this.imagePool.getOrCreateImage(img).then(function () {
              // set dirty rectangle flag
              object.getEntity().getComponent(Renderable).dirty = true;
              renderingService.dirtify();
            });
          }
        }
      });
      renderingService.hooks.attributeChanged.tap(LoadImagePlugin_1.tag, function (object, name, value) {
        var nodeName = object.nodeName;

        if (nodeName === SHAPE.Image) {
          if (name === 'img') {
            if (isString(value)) {
              _this.imagePool.getOrCreateImage(value).then(function () {
                // set dirty rectangle flag
                object.getEntity().getComponent(Renderable).dirty = true;
                renderingService.dirtify();
              });
            }
          }
        }
      });
    }
  }]);

  return LoadImagePlugin;
}(), _class$1.tag = 'LoadImagePlugin', _temp$1);

__decorate([inject(ImagePool), __metadata("design:type", ImagePool)], LoadImagePlugin.prototype, "imagePool", void 0);

LoadImagePlugin = LoadImagePlugin_1 = __decorate([injectable()], LoadImagePlugin);

world.registerComponent(RBushNode);
/**
 * register shape renderers
 */

container.bind(PathGeneratorFactory).toFactory(function (ctx) {
  return function (tagName) {
    if (tagName === SHAPE.Circle) {
      return generatePath;
    } else if (tagName === SHAPE.Ellipse) {
      return generatePath$1;
    } else if (tagName === SHAPE.Rect) {
      return generatePath$2;
    } else if (tagName === SHAPE.Line) {
      return generatePath$3;
    } else if (tagName === SHAPE.Polyline) {
      return generatePath$4;
    } else if (tagName === SHAPE.Polygon) {
      return generatePath$5;
    } else if (tagName === SHAPE.Path) {
      return generatePath$6;
    }

    return null;
  };
});
var containerModule = new ContainerModule(function (bind, unbind, isBound, rebind) {
  bind(ImagePool).toSelf().inSingletonScope();
  bind(RBushRoot).toConstantValue(new RBush());
  bind(DefaultRenderer).toSelf().inSingletonScope();
  bind(ImageRenderer).toSelf().inSingletonScope();
  bind(TextRenderer).toSelf().inSingletonScope();
  bind(StyleRendererFactory).toFactory(function (ctx) {
    return function (tagName) {
      if (tagName === SHAPE.Circle || tagName === SHAPE.Ellipse || tagName === SHAPE.Rect || tagName === SHAPE.Line || tagName === SHAPE.Polyline || tagName === SHAPE.Polygon || tagName === SHAPE.Path) {
        return ctx.container.get(DefaultRenderer);
      } else if (tagName === SHAPE.Image) {
        return ctx.container.get(ImageRenderer);
      } else if (tagName === SHAPE.Text) {
        return ctx.container.get(TextRenderer);
      }

      return null;
    };
  });
  bind(StyleParser).toSelf().inSingletonScope();
  bind(CanvasRendererPlugin).toSelf().inSingletonScope();
  bind(RenderingPluginContribution).toService(CanvasRendererPlugin);
  bind(LoadImagePlugin).toSelf().inSingletonScope();
  bind(RenderingPluginContribution).toService(LoadImagePlugin);
});

export { PathGenerator, PathGeneratorFactory, RBushNode, RBushRoot, containerModule };
